import { ViewPlugin, EditorView, Decoration } from '@codemirror/view';
import { StateEffect, RangeSet, RangeSetBuilder, Compartment, Facet, combineConfig } from '@codemirror/state';
import { initShikiInternal, createTheme, StyleModule, mountStyles } from '@cmshiki/utils';
import { codeToTokens, stringifyTokenStyle, getTokenStyleObject } from '@shikijs/core';

var __defProp$2 = Object.defineProperty;
var __defNormalProp$2 = (obj, key, value) => key in obj ? __defProp$2(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField$2 = (obj, key, value) => {
  __defNormalProp$2(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
const updateEffect = StateEffect.define();
class ShikiView {
  constructor(shikiHighlighter) {
    this.shikiHighlighter = shikiHighlighter;
    __publicField$2(this, "decorations", RangeSet.empty);
    __publicField$2(this, "lastPos", {
      from: 0,
      to: 0
    });
  }
  // when crashed
  destroy() {
    this.clearDecorations();
  }
  update(update) {
    if (update.docChanged) {
      console.log("docChanged");
      this.docChangeHighlight(update);
      return;
    }
    if (update.viewportChanged) {
      this.updateHighlight(update);
      return;
    }
    for (let tr of update.transactions) {
      for (let effect of tr.effects) {
        if (effect.is(updateEffect)) {
          this.shikiHighlighter.update(effect.value, update.view).then(
            () => this.updateHighlight(update)
          );
        }
      }
    }
  }
  clearDecorations() {
    this.decorations = RangeSet.empty;
  }
  docChangeHighlight(update) {
    const { doc } = update.state;
    const newVisibleRanges = update.view.visibleRanges;
    const builder = new RangeSetBuilder();
    if (doc.length === 0) {
      return builder.finish();
    }
    for (let { from, to } of newVisibleRanges) {
      console.log(update);
      update.changes.iterChanges((oldStart, oldEnd, newStart, newEnd, inserted) => {
        if (oldStart === 0 && oldStart === newStart) {
          this.clearDecorations();
          this.updateHighlight(update);
          return;
        }
        this.shikiHighlighter.highlight(doc, from, to, (from2, to2, mark) => {
          builder.add(from2, to2, mark);
        });
        this.lastPos.from = from;
        this.lastPos.to = to;
        this.decorations = builder.finish();
      });
    }
  }
  updateHighlight(update) {
    const builder = new RangeSetBuilder();
    const doc = update.state.doc;
    const newVisibleRanges = update.view.visibleRanges;
    for (let { from, to } of newVisibleRanges) {
      this.shikiHighlighter.highlight(doc, from, to, (from2, to2, mark) => {
        builder.add(from2, to2, mark);
      });
      this.lastPos = { from, to };
    }
    this.decorations = builder.finish();
  }
}
const shikiViewPlugin = (shikiHighlighter, _options) => {
  return {
    viewPlugin: ViewPlugin.define(
      (view) => new ShikiView(shikiHighlighter.setView(view)),
      {
        decorations: (v) => v.decorations
      }
    )
  };
};

var __defProp$1 = Object.defineProperty;
var __defNormalProp$1 = (obj, key, value) => key in obj ? __defProp$1(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField$1 = (obj, key, value) => {
  __defNormalProp$1(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
const themeCompartment = new Compartment();
const configsFacet = Facet.define({
  combine: (values) => combineConfig(values, {}, {
    themeStyle: (f, s) => s
  })
});
class Base {
  constructor(shikiCore, configs) {
    this.shikiCore = shikiCore;
    this.configs = configs;
    __publicField$1(this, "themesCache", /* @__PURE__ */ new Map());
    __publicField$1(this, "currentTheme", "light");
    this.loadThemes();
  }
  /** determines whether the theme style of the current option is `cm` or not */
  get isCmStyle() {
    return this.configs.themeStyle === "cm";
  }
  static init(shikiCore, configs) {
    return new Base(shikiCore, configs);
  }
  /** 
   * get default theme
   */
  initTheme() {
    const { defaultColor } = this.configs;
    if (defaultColor === false) {
      return EditorView.baseTheme({});
    }
    this.currentTheme = defaultColor || "light";
    return this.getTheme(this.currentTheme);
  }
  async update(options, view) {
    if (options.theme) {
      options.themes = options.themes || {
        ...this.configs.themes,
        light: options.theme
      };
    }
    const _configs = {
      ...this.configs
    };
    this.configs = {
      ...this.configs,
      ...options
    };
    if (JSON.stringify(options.themes) !== JSON.stringify(_configs.themes) || _configs.defaultColor !== options.defaultColor || _configs.lang !== options.lang || _configs.langAlias !== options.langAlias || _configs.warnings !== options.warnings) {
      this.shikiCore = await initShikiInternal(this.configs);
      window.requestAnimationFrame(() => {
        this.loadThemes();
        view.dispatch({
          effects: themeCompartment.reconfigure(this.initTheme())
        });
      });
    }
  }
  /**
   * get theme
   *
   * @param {string} name `light\dark\...`
   * @returns {Extension} codemirror theme extension
   * @throws `xxxx theme not registered!`
   */
  getTheme(name = this.currentTheme) {
    if (this.themesCache.get(name)) {
      return this.themesCache.get(name);
    } else {
      throw new Error(`'${name}' theme is not registered!`);
    }
  }
  /**
   * preload all registered themes and create codemirror theme extension
   */
  loadThemes() {
    let { themes, cssVariablePrefix } = this.configs;
    Object.keys(themes).forEach((color) => {
      const name = themes[color];
      if (!name)
        throw new Error(`'${name}' theme is not registered!`);
      const { colors, bg, fg, type } = this.shikiCore.getTheme(name);
      const prefix = cssVariablePrefix + color;
      let settings = {
        background: bg,
        foreground: fg
      };
      if (colors) {
        settings = {
          ...settings,
          gutterBackground: bg,
          gutterForeground: fg,
          // fontFamily
          // fontSize
          // gutterActiveForeground
          gutterBorder: "transparent",
          selection: colors["editor.wordHighlightBackground"] || colors["editor.selectionBackground"],
          selectionMatch: colors["editor.wordHighlightStrongBackground"] || colors["editor.selectionBackground"],
          caret: colors["editorCursor.foreground"] || colors["foreground"] || "#FFFFFF",
          // dropdownBackground: colors['dropdown.background'],
          // dropdownBorder: colors['dropdown.border'] || colors['foreground'],
          lineHighlight: colors["editor.lineHighlightBackground"] || "#50505044"
          // matchingBracket: colors['editorBracketMatch.background'] || colors['editor.lineHighlightBackground'] || colors['editor.selectionBackground'],
        };
      }
      const extension = createTheme({
        theme: type,
        settings,
        classes: {
          [`& .cm-line span`]: {
            color: `var(${prefix}) !important`,
            // backgroundColor: `var(${prefix}-bg) !important`,
            fontStyle: `var(${prefix}-font-style) !important`,
            fontWeight: `var(${prefix}-font-weight) !important`,
            textDecoration: `var(${prefix}-text-decoration) !important`
          }
        }
      });
      this.themesCache.set(color, extension);
    });
  }
}

function toStyleObject(styleStr, isBgStyle = false) {
  const Styles = {};
  styleStr.split(";").forEach((str, i) => {
    const [k, v] = str.split(":");
    if (k && v) {
      Styles[k] = v;
    } else {
      if (isBgStyle) {
        Styles["background-color"] = k;
      } else {
        Styles["color"] = k;
      }
    }
  });
  return Styles;
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class ShikiHighlighter extends Base {
  constructor(shikiCore, options) {
    super(shikiCore, options);
    __publicField(this, "view");
  }
  setView(view) {
    this.view = view;
    return this;
  }
  // getLastGrammarState(preCode: string) {
  // }
  codeToTokens(code) {
    return codeToTokens(this.shikiCore, code, this.configs);
  }
  /**
  * add highlighting to text
  *
  * @param doc content text
  * @param from text start
  * @param to text end
  * @returns `{ decorations }` an object that contains decorative information
  */
  highlight(doc, from, to, buildDeco, preState) {
    const { cssVariablePrefix, defaultColor } = this.configs;
    const content = doc.sliceString(from, to);
    const { tokens } = this.codeToTokens(content);
    let pos = from;
    let cmClasses = {};
    this.view.dom.classList.toggle("lang-" + this.configs.lang);
    tokens.forEach((lines) => {
      lines.forEach((token) => {
        let style = (token.htmlStyle || stringifyTokenStyle(getTokenStyleObject(token))).replace(/color/g, cssVariablePrefix + defaultColor);
        ["font-style", "font-weight", "text-decoration"].forEach((s) => {
          style = style.replace(new RegExp(`;` + s, "g"), `;${cssVariablePrefix + defaultColor}-${s}`);
        });
        cmClasses[style] = cmClasses[style] || StyleModule.newName();
        let to2 = pos + token.content.length;
        buildDeco(
          pos,
          to2,
          Decoration.mark({
            tagName: "span",
            attributes: {
              [this.isCmStyle ? "class" : "style"]: this.isCmStyle ? cmClasses[style] : style
            }
          })
        );
        pos = to2;
      });
      pos++;
    });
    if (this.isCmStyle) {
      Object.entries(cmClasses).forEach(([k, v]) => {
        mountStyles(this.view, {
          [`& .cm-line .${v}`]: toStyleObject(k)
        });
      });
    }
  }
}

const shikiPlugin = async (highlighter, ctOptions) => {
  const shikiHighlighter = new ShikiHighlighter(
    highlighter,
    ctOptions
  );
  const { viewPlugin } = shikiViewPlugin(shikiHighlighter);
  const BaseCore = Base.init(highlighter, ctOptions);
  return {
    /**
    * get theme
    *
    * @param { string } name `light\dark\...`
    * @returns { Extension } codemirror theme extension
    * @throws `xxxx theme not registered!`
    */
    getTheme(name, view) {
      return view ? shikiHighlighter.setView(view).getTheme(name) : BaseCore.getTheme(name);
    },
    shiki: [
      themeCompartment.of(BaseCore.initTheme()),
      viewPlugin
    ]
  };
};

const defaultOptions = {
  lang: "text",
  warnings: true,
  themeStyle: "cm",
  defaultColor: "light",
  cssVariablePrefix: "--shiki-",
  tokenizeMaxLineLength: 2e4,
  includeExplanation: false,
  tokenizeTimeLimit: 500
};
const defaultOptions$1 = defaultOptions;

async function shikiToCodeMirror(shikiOptions) {
  const { theme, themes } = shikiOptions;
  if (!themes) {
    if (theme) {
      shikiOptions.themes = {
        light: theme
      };
    } else {
      throw new Error("[@cmshiki/shiki]Invalid options, either `theme` or `themes` must be provided");
    }
  }
  const options = {
    ...defaultOptions$1,
    ...shikiOptions
  };
  const shikiInternalCore = await initShikiInternal(options);
  return shikiPlugin(shikiInternalCore, options);
}

export { configsFacet, shikiToCodeMirror, themeCompartment, updateEffect };
